%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This function reconstructs the trajectory from sampled data points by 
% implementing the algorithm for Trajectory Reconstruction using a LINEAR 
% Generative Model by Dey & Krishnaprasad, presented in 50th Annual 
% Allerton Conference - 2012, Monticello, IL , USA.
%
% Written and compiled by Biswadip Dey, University of Maryland in May 2012.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% **************************** Initialization ****************************
%
function ...
    [Fit_Error, Fine_POSITION, Fine_VELOCITY, Fine_ACCELRTN, Fine_JERK, ...
    Reconstructed_CURVE, Curve, Fine_TIME] ...
    = Linear_Filter(Lambda,Sub_Interval_Res,Curve,Time_Array,FrameRate)
%
% ****** Lambda:>>> Regularization Parameter
% ****** Curve:>>> Data is Arranged HORIZONTALLY
% ****** Time_Array:>>> Time Tags (They are basically serial numbers for 
%                       the snapshots)
% ****** FrameRate:>>> Image capture rate
% ****** Sub_Interval_Res:>>> Number of points to be inserted between two 
%                             Data-points
%%
%
Inc_Time_Array = diff(Time_Array)*(1/FrameRate);
%
%%
% A,B,C ---->>> Matrices defining the system dynamics
A = [zeros(3) eye(3) zeros(3); ...
    zeros(3) zeros(3) eye(3); ...
    zeros(3) zeros(3) zeros(3)];
%
B = [zeros(3) zeros(3) eye(3)]';
%
C = [eye(3) zeros(3) zeros(3)];
%
%
% ---------- Creating the State Transition Matrices ----------
%
Discrete_Dynamics_UPSILON = cell(1,length(Inc_Time_Array));
for time_idx = 1:length(Inc_Time_Array)
    %
    %
    t_delta = Inc_Time_Array(time_idx);
    Grammian_W = ...
        [ (t_delta^5/20)*eye(3) (-t_delta^4/8)*eye(3) (t_delta^3/6)*eye(3);...
        (-t_delta^4/8)*eye(3) (t_delta^3/3)*eye(3)  (-t_delta^2/2)*eye(3);...
        (t_delta^3/6)*eye(3)  (-t_delta^2/2)*eye(3) (t_delta)*eye(3)];
    %
    SubMatrix_11 = expm(A*t_delta);
    SubMatrix_12 = -expm(A*t_delta)*Grammian_W;
    %
    SubMatrix_21 = zeros(size(A));
    SubMatrix_22 = expm(-A'*t_delta);
    %
    Discrete_Dynamics_UPSILON{time_idx} = ...
        [SubMatrix_11 SubMatrix_12; SubMatrix_21 SubMatrix_22];
    %
    %
end
%
Discrete_Dynamics_M = [eye(size(A)) zeros(size(A)); ...
    (-1/Lambda)*(C'*C) eye(size(A))];
%
%
%
% ---------- Creating the i/p matrix through which Data affects ----------
Discrete_Dynamics_GAMMA = [zeros(size(C')); (1/Lambda)*C'];
%
%
%% --------------------------- Solving x(t_0) ---------------------------
%
%
Left_Product_LAMBDA = cell(1,length(Inc_Time_Array));
Sum_of_Product_Data = cell(1,length(Inc_Time_Array));
%
Left_Product_LAMBDA{1} = Discrete_Dynamics_M*Discrete_Dynamics_UPSILON{1};
Sum_of_Product_Data{1} = ...
    Discrete_Dynamics_M*Discrete_Dynamics_UPSILON{1}*Discrete_Dynamics_GAMMA*Curve(:,1) ...
    + Discrete_Dynamics_GAMMA*Curve(:,2);
%
%
for time_idx = 2:length(Inc_Time_Array)
    %
    Left_Product_LAMBDA{time_idx} = ...
        Discrete_Dynamics_M*Discrete_Dynamics_UPSILON{time_idx}*Left_Product_LAMBDA{time_idx-1};
    %
    Sum_of_Product_Data{time_idx} = ...
        Discrete_Dynamics_M*Discrete_Dynamics_UPSILON{time_idx}*Sum_of_Product_Data{time_idx-1}...
        + Discrete_Dynamics_GAMMA*Curve(:,time_idx+1);
end
%
%
%
LHS_of_InitialValue_equation = [zeros(size(A)) eye(size(A))]*Left_Product_LAMBDA{length(Inc_Time_Array)}*[eye(size(A)); -(1/Lambda)*(C'*C)];
%
RHS_of_InitialValue_equation = [zeros(size(A)) -eye(size(A))]*Sum_of_Product_Data{length(Inc_Time_Array)};
%
Initial_STATE = LHS_of_InitialValue_equation\RHS_of_InitialValue_equation;
%
%
%% ------------------------- State Reconstruction -------------------------
%
%
Reconstructed_State = NaN(18,length(Time_Array));
%
%
Reconstructed_State(:,1) = ...
    [eye(size(A)); -(1/Lambda)*(C'*C)]*Initial_STATE ...
    + Discrete_Dynamics_GAMMA*Curve(:,1);
%
%
for time_idx = 2:length(Time_Array)
    %
    Reconstructed_State(:,time_idx) = Left_Product_LAMBDA{time_idx-1}*[eye(size(A)); -(1/Lambda)*(C'*C)]*Initial_STATE + Sum_of_Product_Data{time_idx-1};
end
%
% ---------------- Reconstructed Curve ----------------
Reconstructed_CURVE = C*[eye(size(A)) zeros(size(A))]*Reconstructed_State;
%
%
%% ------------------------- Performance Measure -------------------------
%
%
MisMatch_Reconstruction = Curve - Reconstructed_CURVE;
Fit_Error = sum(sum(MisMatch_Reconstruction.^2));
%
%
%toc
%% ------------------------ Increasing Resolution ------------------------
%
%
length_fine_info = 1 + (max(Time_Array)-min(Time_Array))*(Sub_Interval_Res+1);
Fine_POSITION = NaN(3,length_fine_info);
Fine_VELOCITY = NaN(3,length_fine_info);
Fine_ACCELRTN = NaN(3,length_fine_info);
%
Fine_JERK = NaN(3,length_fine_info);
%
Time_idx_finer = Time_Array(1):1/(1+Sub_Interval_Res):Time_Array(length(Time_Array));
Fine_TIME = (Time_idx_finer-Time_idx_finer(1))*(1/FrameRate);
%
%
C_vel = [zeros(3) eye(3) zeros(3)];
C_acl = [zeros(3) zeros(3) eye(3)];
%
%
Absolute_idx = 0;
for time_idx = 1:length(Time_Array)-1
    %
    Absolute_idx = Absolute_idx + 1;
    abs_idx_for_sample_time = Absolute_idx;
    %
    Fine_POSITION(:,Absolute_idx) = C*[eye(size(A)) zeros(size(A))]*Reconstructed_State(:,time_idx);
    Fine_VELOCITY(:,Absolute_idx) = C_vel*[eye(size(A)) zeros(size(A))]*Reconstructed_State(:,time_idx);
    Fine_ACCELRTN(:,Absolute_idx) = C_acl*[eye(size(A)) zeros(size(A))]*Reconstructed_State(:,time_idx);
    %
    Fine_JERK(:,Absolute_idx) = -B'*[zeros(size(A)) eye(size(A))]*Reconstructed_State(:,time_idx);
    %
    %
    diff_from_next = Time_Array(time_idx+1)-Time_Array(time_idx);
    %
    for sub_point = 1:(1+Sub_Interval_Res)*diff_from_next - 1
        %
        Absolute_idx = Absolute_idx + 1;
        %
        incremental_time = Fine_TIME(Absolute_idx) - Fine_TIME(abs_idx_for_sample_time);
        %
        Gram_Matrix = ...
            [ (incremental_time^5/20)*eye(3) (-incremental_time^4/8)*eye(3) (incremental_time^3/6)*eye(3);...
            (-incremental_time^4/8)*eye(3) (incremental_time^3/3)*eye(3)  (-incremental_time^2/2)*eye(3);...
            (incremental_time^3/6)*eye(3)  (-incremental_time^2/2)*eye(3) (incremental_time)*eye(3)];
        %
        temp_Recon_Traj_all = expm(A*incremental_time)*[eye(size(A)) -Gram_Matrix]*Reconstructed_State(:,time_idx);
        %
        %
        Fine_POSITION(:,Absolute_idx) = C*temp_Recon_Traj_all;
        Fine_VELOCITY(:,Absolute_idx) = C_vel*temp_Recon_Traj_all;
        Fine_ACCELRTN(:,Absolute_idx) = C_acl*temp_Recon_Traj_all;
        %
        Fine_JERK(:,Absolute_idx) = -B'*expm(-A'*incremental_time)*[zeros(size(A)) eye(size(A))]*Reconstructed_State(:,time_idx);
        %
    end
end
%
Absolute_idx = Absolute_idx + 1;
%
Fine_POSITION(:,Absolute_idx) = C*[eye(size(A)) zeros(size(A))]*Reconstructed_State(:,length(Time_Array));
Fine_VELOCITY(:,Absolute_idx) = C_vel*[eye(size(A)) zeros(size(A))]*Reconstructed_State(:,length(Time_Array));
Fine_ACCELRTN(:,Absolute_idx) = C_acl*[eye(size(A)) zeros(size(A))]*Reconstructed_State(:,length(Time_Array));
%
Fine_JERK(:,Absolute_idx) = -B'*[zeros(size(A)) eye(size(A))]*Reconstructed_State(:,length(Time_Array));
%
Fine_TIME(Absolute_idx) = (Time_Array(length(Time_Array)) - Time_Array(1))/FrameRate;