% This is an implementation of the algorithm for Trajectory Reconstruction
% using a LINEAR Generative Model by Dey & Krishnaprasad, presented in 50th
% Annual Allerton Conference - 2012, Monticello, IL , USA.
%
% This algorithm reconstructs a trajectory from sampled data by penalizing
% high values of the jerk path integral. As this is a regularized inversion
% problem, the value of the regularization parameter (lambda) plays an
% important role [refer our paper]. We choose the optimal value for the
% regularization parameter by applying leave-one-out approach for
% ordinary cross validation (OCV).
% This algorithm provides all the relevant parameters of motion (namely,
% speed, classical and natural curvatures, torsion, velocity, acceleration
% and lateral acceleration).
%
%
% Written and compiled by Biswadip Dey, University of Maryland in May 2012.
%
% If you use this code please refer the following paper:
% "Trajectory Smoothing As a Linear Optimal Control Problem";
% Biswadip Dey, P. S. Krishnaprasad;
% 50th Annual Allerton Conference on Communication, Control, and Computing,
% October, 2012, Monticello, IL, USA
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% NOTE: This is moderately fast, depending on the number of sampled data
% points and the size of Lambda of values over which we perform the cross
% validation.
%
% NOTE: contact information - biswadipdey.1984@gmail.com
%
% NOTE: As the cross validation process for computing optimal value for the
% regularization parameter is parallel in nature, we extract benefit from
% multi-core processor and Matlab Parallel Computing Toolbox (PCT). If you
% are equipped with a multi-core computer and Matlab Parallel Computing
% Toolbox, then by modifying (commenting and uncommenting) a couple of 
% lines in the "Ord_Crss_Validation.m" file you can make the program 
% faster. The factor of reduction in execution time is approximately same 
% as the number of available cores in your computer. The details about this 
% modification is provided within the "Ord_Crss_Validation.m" file.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% **************************** Initialization ****************************
%
%
function [Return_4rm_Recon] = ...
    MAIN_code_4_Reconstruction(Raw_Data,Time_IDX,LAMBDA_ARRAY,FrameRate,Pts_2B_Insrtd)
%
%****** Raw_Data >>>>>> Data is Arranged HORIZONTALLY. (is a 3xN matrix, 
%                       N is the number of data points)
%****** Time_IDX >>>>>> Time Tags (They are basically serial numbers for 
%                       the snapshots, and is used to keep a track of
%                       missing frames, if any.)
%****** FrameRate >>>>>> Image capture rate (fps)
%****** LAMBDA_ARRAY >>>>>> Stores the values for the regularizing
%                           parameter for which we compute the cross 
%                           validation cost.
%****** Pts_2B_Insrtd >>>>>> Number of data points that a user wants to
%                            insert between two data ponits.
%%
%
clc;
echo off;
warning off all
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ******************************** OCV ***********************************
%
display('Beginning Cross Validation ....') 
%
% Performs OCV.
[L_ArRaY,Err_MaTrIx,OCV_ErRoR] = Ord_Crss_Validation(Raw_Data,Time_IDX,FrameRate,LAMBDA_ARRAY);
%
% Computes the optimal value of the regularization parameter by picking up
% the "lambda"-value which minimizes the Cross Validation Error.
[~,idx_4_L_opt] = min(OCV_ErRoR);
L_star = L_ArRaY(idx_4_L_opt);
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ************************ Reconstruction ********************************
%
%
% Reconstructs the Trajectory using the optimal value of the regularization 
% parameter.
[Error_FiT, PoSiTiOn, VeLoCiTy, AcCeLeRaTiOn, ...
    JeRk, recon_CuRvE, orig_CuRvE, TIME_fine] ...
    = Linear_Filter(L_star,Pts_2B_Insrtd,Raw_Data,Time_IDX,FrameRate);
%
% Computes the jerk path intergral using trapezoidal rule for numerical 
% integration.
avg_jrk = 0.5*(JeRk(:,1:size(JeRk,2)-1)+JeRk(:,2:size(JeRk,2)));
Error_smooth_JrK = L_star*sum(sum(avg_jrk.^2).*diff(TIME_fine));
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ********************* Parameters of Motion *****************************
%
%
% Extracts different metrics of interest from the results of
% reconstruction.
%
SpEeD = (sqrt(sum(VeLoCiTy.^2)));
% Magnitude of the velocity vector
%
unit_TaNgEnT = bsxfun(@rdivide,VeLoCiTy,SpEeD);
% Normalized velocity
%
fwd_ThRuSt = (sum(AcCeLeRaTiOn.*unit_TaNgEnT));
% Forward thrust is defined as the component of acceleration along the
% direction of velocity. Alternatively, it can interpreted as the
% derivative of speed.
%
T_dot = bsxfun(@rdivide,...
    (AcCeLeRaTiOn - bsxfun(@times,unit_TaNgEnT,fwd_ThRuSt)),...
    SpEeD);
% This parameter can be interpreted as the lateral acceleration (normalized
% by speed).
%
classical_CuRvAtUrE = sqrt(sum(T_dot.^2))./SpEeD;
% Classical curvature (denoted by "\kappa" in literature) can be
% interpreted as the turning rate of the curve.
%
ToRsIoN = NaN(size(classical_CuRvAtUrE)); % \tau
ThEtA = zeros(size(classical_CuRvAtUrE));
for i = 1:size(classical_CuRvAtUrE,2) - 1
    %
    Temp_matrix = [VeLoCiTy(:,i) AcCeLeRaTiOn(:,i) JeRk(:,i)];
    ToRsIoN(i) = (det(Temp_matrix))/((norm(cross(VeLoCiTy(:,i),AcCeLeRaTiOn(:,i))))^2);
    %
    ThEtA(i+1) = ThEtA(i) + (TIME_fine(i+1)-TIME_fine(i))*ToRsIoN(i);
end
% Theta can be interpreted as the accumulated torsion.
%
nat_CuRvAtUrE_U = classical_CuRvAtUrE.*cos(ThEtA);
nat_CuRvAtUrE_V = classical_CuRvAtUrE.*sin(ThEtA);
% The natural curvatures are related to the classical curvature in a way
% similar to cartesian to polar co-ordinate transformation.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ******************** Packaging the Results *****************************
%
%
Return_4rm_Recon = struct('OCV_Error_Grid',Err_MaTrIx,...
    'Cross_Validation_Error',OCV_ErRoR,'optimal_lambda',L_star,...
    'PoSiTiOn',PoSiTiOn,'VeLoCiTy',VeLoCiTy,'AcCeLeRaTiOn',AcCeLeRaTiOn,...
    'JeRk',JeRk,...
    'recon_CuRvE',recon_CuRvE,'orig_CuRvE',orig_CuRvE,...
    'TIME_fine',TIME_fine','Error_FiT',Error_FiT,...
    'JerK_Path_INT',Error_smooth_JrK,...
    'SpEeD',SpEeD,'unit_TaNgEnT',unit_TaNgEnT,'fwd_ThRuSt',fwd_ThRuSt,...
    'Lat_accln',T_dot,...
    'classical_CuRvAtUrE',classical_CuRvAtUrE,...
    'ToRsIoN',ToRsIoN,'ThEtA',ThEtA,...
    'nat_CuRvAtUrE_U',nat_CuRvAtUrE_U,'nat_CuRvAtUrE_V',nat_CuRvAtUrE_V);
% *********************************************************************